{
+----------------------------------------------------------------------------+
|                                                                          |
|                                                                       |
|                                                                      |
|                                                                       |
|                                                                       |
|                                                                    |
|                                                             |
|                                                        |
|                                                      |
|                       Copyright  1996-1997 by:  |
|                                                  |
|                           WHITE ANTS SYSTEMHOUSE BV  |
|                            Geleen 12                  |
|                                  8032 GB Zwolle             |
|                                        Netherlands                |
|                                                               |
|                                         Tel. +31 38 453 86 31      |
|                                              Fax. +31 38 453 41 22      |
|                                                                        |
|                                             www.whiteants.com          |
|                                            support@whiteants.com      |
|                                                                           |
+----------------------------------------------------------------------------+
  file     : Timers
  version  : 1.0
  comment  :
  date     : 14-01-97
  time     : 
  author   : R. Post
  compiler : Delphi 1.0
+----------------------------------------------------------------------------+
| DISCLAIMER:                                                                |
| THIS SOURCE IS FREEWARE. YOU ARE ALLOWED TO USE IT IN YOUR OWN PROJECTS    |
| WITHOUT ANY RESTRICTIONS. YOU ARE NOT ALLOWED TO SELL THE SOURCE CODE.     |
| THERE IS NO WARRANTY AT ALL - YOU USE IT ON YOUR OWN RISC. WHITE ANTS DOES |
| NOT ASSUME ANY RESPONSIBILITY FOR ANY DAMAGE OR ANY LOSS OF TIME OR MONEY  |
| DUE THE USE OF ANY PART OF THIS SOURCE CODE.                               |
+----------------------------------------------------------------------------+
}

unit Timers;

{.$DEFINE RAISE_EXCEPTIONS}

interface

uses
  WinTypes, WinProcs, SysUtils, Classes, ExtCtrls,
  Filters, Containr;

const
  SUnableToRemoveFmt = 'Unable to remove timer %d from TimerPool';

type
  ETimerError = class(Exception);

  TTriggerEvent = procedure (Sender: TObject; Handle: Integer) of object;

  TPoolTimer = class (TObject)
  private
    FHandle: Integer;
    FIntervalMS: LongInt;
    FOnTrigger: TTriggerEvent;
    FStarted: Boolean;
    FStartTimeMS: LongInt;
    FTerminated: Boolean;
  protected
    procedure Trigger;
  public
    function Expired: Boolean;
    procedure FireOnExpire; virtual; abstract;
    procedure Restart;
    procedure Start;
    procedure Stop;
    procedure Terminate;
    property Handle: Integer read FHandle write FHandle;
    property IntervalMS: LongInt read FIntervalMS write FIntervalMS;
    property OnTrigger: TTriggerEvent read FOnTrigger write FOnTrigger;
    property Started: Boolean read FStarted;
    property StartTimeMS: LongInt read FStartTimeMS;
    property Terminated: Boolean read FTerminated;
  end;

  TIntervalPoolTimer = class (TPoolTimer)
  public
    procedure FireOnExpire; override;
  end;

  TOneShotPoolTimer = class (TPoolTimer)
  public
    procedure FireOnExpire; override;
  end;

  TDelayPoolTimer = class (TPoolTimer)
  public
    procedure FireOnExpire; override;
  end;

  TPoolTimerCollection = class (TCollection)
  protected
    function GetPoolTimers(Index: Integer): TPoolTimer;
  public
    constructor Create;
    function Compare(Key1, Key2: Pointer): Integer; override;
    property PoolTimers[Index: Integer]: TPoolTimer read GetPoolTimers;
  end;

  TTimerPool = class (TObject)
  private
    FLastHandle: Integer;
    FPoolTimers: TPoolTimerCollection;
    FTimer: TTimer;
  protected
    FPoolTimerKey: TPoolTimer;
    function AllocateHandle: Integer;
    function GetCount: Integer;
    function GetStarted(AHandle: Integer): Boolean;
    procedure TimerTimer(Sender: TObject);
  public
    constructor Create;
    destructor Destroy; override;
    function Add(ATriggerProc: TTriggerEvent; AInterval: LongInt): Integer;
    function AddDelay(ATriggerProc: TTriggerEvent; AInterval: LongInt): Integer;
    function AddOneTime(ATriggerProc: TTriggerEvent; AInterval: LongInt): 
        Integer;
    procedure Remove(AHandle: Integer);
    procedure RemoveAll;
    procedure Restart(AHandle: Integer);
    property Count: Integer read GetCount;
    property LastHandle: Integer read FLastHandle;
    property Started[AHandle: Integer]: Boolean read GetStarted;
  end;

  TTimerPoolComponent = class (TComponent)
  private
    FEnabled: Boolean;
    FHandle: Integer;
    FInterval: LongInt;
    FOnTimer: TNotifyEvent;
  protected
    function GetTimerPool: TTimerPool;
    procedure SetEnabled(Value: Boolean);
    procedure SetInterval(Value: LongInt);
    procedure Timer;
    procedure TimerPoolTrigger(Sender: TObject; Handle: Integer); virtual;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure UpdateTimer; virtual; abstract;
    property Handle: Integer read FHandle write FHandle;
    property TimerPool: TTimerPool read GetTimerPool;
  published
    property Enabled: Boolean read FEnabled write SetEnabled;
    property Interval: LongInt read FInterval write SetInterval;
    property OnTimer: TNotifyEvent read FOnTimer write FOnTimer;
  end;

  TIntervalTimer = class (TTimerPoolComponent)
  public
    procedure UpdateTimer; override;
  end;

  TOneShotTimer = class (TTimerPoolComponent)
  public
    procedure UpdateTimer; override;
  end;

  TDelayTimer = class (TTimerPoolComponent)
  public
    procedure UpdateTimer; override;
  end;


{ Truncates the value of the passed double to an unsigned longint value and
  returns this as a longint. }
function UTrunc(aDouble: Double): LongInt;

{ Treats the value of the passed longint as an unsigned longint and
  returns the unsigned value as a double. }

function ULong(L: LongInt): Double;

{ The AccurateTimeMS function retrieves the number of milliseconds that have
  elapsed since Windows was started. In enhanced mode it is accurate to 1 ms.
  If windows has been active for more then 24.5 days, the LongInt value will
  go negative and after 49 days,the timer will be overflow to 0. }
function AccurateTimeMS: LongInt;

{ Calculate time elapsed (in milliseconds) between Start and Stop.
  Start and Stop should be values retrieved with AccurateTimeMS or GetTickCount.
  This function can determine the elapsed time up to 49 days. }
function ElapsedTimeMS(Start, Stop : LongInt): Double;

function EncodeTimeMS(Hour, Min, Sec, MSec: Word): LongInt;

procedure DecodeTimeMS(TimeMS: LongInt; var Hour, Min, Sec, MSec: Word);

var
  TimerPool: TTimerPool;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('White Ants', [TOneShotTimer, TIntervalTimer, TDelayTimer]);
end;


{
********************************** TPoolTimer **********************************
}
function TPoolTimer.Expired: Boolean;
begin
  Result := False;
  if Started then
    if ElapsedTimeMS( FStartTimeMS, AccurateTimeMS) > FIntervalMS then
      Result := True;
end;

procedure TPoolTimer.Restart;
begin
  if not Terminated then
  begin
    FStartTimeMS := AccurateTimeMS;
    FStarted := True;
  end;
end;

procedure TPoolTimer.Start;
begin
  if not Started then Restart;
end;

procedure TPoolTimer.Stop;
begin
  FStarted := False;
end;

procedure TPoolTimer.Terminate;
begin
  FTerminated := True;
end;

procedure TPoolTimer.Trigger;
begin
  if Assigned(FOnTrigger) then FOnTrigger(Self, FHandle);
end;

{
****************************** TIntervalPoolTimer ******************************
}
procedure TIntervalPoolTimer.FireOnExpire;
begin
  if Expired then
  begin
    Restart;
    Trigger;
  end;
end;

{
****************************** TOneShotPoolTimer ******************************
}
procedure TOneShotPoolTimer.FireOnExpire;
begin
  if Expired then
  begin
    Stop;
    try
      Trigger;
    finally
      Terminate;
    end;
  end;
end;

{
******************************* TDelayPoolTimer *******************************
}
procedure TDelayPoolTimer.FireOnExpire;
begin
  if Expired then
  begin
    Stop;
    try
      Trigger;
    finally
      Start;
    end;
  end;
end;

{
***************************** TPoolTimerCollection *****************************
}
constructor TPoolTimerCollection.Create;
begin
  inherited Create;
  CanSort := True;
  Sorted := True;
  OwnesItems := True;
end;

function TPoolTimerCollection.Compare(Key1, Key2: Pointer): Integer;
begin
  if TPoolTimer(Key1).Handle < TPoolTimer(Key2).Handle then
    Result := -1
  else
    if TPoolTimer(Key1).Handle > TPoolTimer(Key2).Handle then
      Result := 1
    else
      Result := 0;
end;

function TPoolTimerCollection.GetPoolTimers(Index: Integer): TPoolTimer;
begin
  Result := Items[Index];
end;

{
********************************** TTimerPool **********************************
}
constructor TTimerPool.Create;
begin
  inherited Create;
  FPoolTimerKey := TPoolTimer.Create;
  FPoolTimers := TPoolTimerCollection.Create;
  FTimer := TTimer.Create( nil);
  FTimer.OnTimer := TimerTimer;
  FTimer.Interval := 50;
  FTimer.Enabled := True;
end;

destructor TTimerPool.Destroy;
begin
  FPoolTimerKey.Free;
  FPoolTimers.Free;
  FTimer.Free;
  inherited Destroy;
end;

function TTimerPool.Add(ATriggerProc: TTriggerEvent; AInterval: LongInt): 
    Integer;
var
  NewPoolTimer: TIntervalPoolTimer;
begin
  { Allocate a new pooltimer. }
  NewPoolTimer := TIntervalPoolTimer.Create;
  NewPoolTimer.Handle := AllocateHandle;
  NewPoolTimer.IntervalMS := AInterval;
  NewPoolTimer.OnTrigger := ATriggerProc;
  
  { Add the PoolTimer to the list. }
  FPoolTimers.Add(NewPoolTimer);
  
  { Return the handle of the new pool timer. }
  Result := NewPoolTimer.Handle;
  
  { Finally start the timer. }
  NewPoolTimer.Start;
end;

function TTimerPool.AddDelay(ATriggerProc: TTriggerEvent; AInterval: LongInt): 
    Integer;
var
  NewPoolTimer: TDelayPoolTimer;
begin
  { Allocate a new pooltimer. }
  NewPoolTimer := TDelayPoolTimer.Create;
  NewPoolTimer.Handle := AllocateHandle;
  NewPoolTimer.IntervalMS := AInterval;
  NewPoolTimer.OnTrigger := ATriggerProc;
  
  { Add the PoolTimer to the list. }
  FPoolTimers.Add(NewPoolTimer);
  
  { Return the handle of the new pool timer. }
  Result := NewPoolTimer.Handle;
  
  { Finally start the timer. }
  NewPoolTimer.Start;
end;

function TTimerPool.AddOneTime(ATriggerProc: TTriggerEvent; AInterval: 
    LongInt): Integer;
var
  NewPoolTimer: TOneShotPoolTimer;
begin
  { Allocate a new pooltimer. }
  NewPoolTimer := TOneShotPoolTimer.Create;
  NewPoolTimer.Handle := AllocateHandle;
  NewPoolTimer.IntervalMS := AInterval;
  NewPoolTimer.OnTrigger := ATriggerProc;
  
  { Add the PoolTimer to the list. }
  FPoolTimers.Add(NewPoolTimer);
  
  { Return the handle of the new pool timer. }
  Result := NewPoolTimer.Handle;
  
  { Finally start the timer. }
  NewPoolTimer.Start;
end;

function TTimerPool.AllocateHandle: Integer;
var
  IdxHandle: Integer;
  MaxIdx: Integer;
  MinIdx: Integer;
  I: Integer;
begin
  { Look for the maximum and minimum indices that are in use. }
  if FPoolTimers.Count = 0 then
  begin
    MinIdx := 10000;
    MaxIdx := 10000;
  end
  else
  begin
    MinIdx := 32000;
    MaxIdx := 0;
    for I := 0 to FPoolTimers.Count-1 do
    begin
      IdxHandle := FPoolTimers.PoolTimers[I].Handle;
      if MaxIdx < IdxHandle then
        MaxIdx := IdxHandle;
      if MinIdx > IdxHandle then
        MinIdx := IdxHandle;
    end;
  end;
  
  if MinIdx > 0 then
    Result := MinIdx-1
  else
    Result := MaxIdx+1;
  
  FLastHandle := Result;
end;

function TTimerPool.GetCount: Integer;
begin
  Result := FPoolTimers.Count;
end;

function TTimerPool.GetStarted(AHandle: Integer): Boolean;
var
  Idx: Integer;
begin
  FPoolTimerKey.Handle := AHandle;
  if FPoolTimers.Find(FPoolTimerKey, Idx) then
    Result := FPoolTimers.PoolTimers[Idx].Started
  else
    Result := False;
end;

procedure TTimerPool.Remove(AHandle: Integer);
var
  Idx: Integer;
begin
  FPoolTimerKey.Handle := AHandle;
  if not FPoolTimers.Find(FPoolTimerKey, Idx) then
  {$IFDEF RAISE_EXCEPTIONS}
    raise ETimerError.CreateFmt(SUnableToRemoveFmt, [AHandle])
  {$ENDIF}
  else
  begin
    FPoolTimers.PoolTimers[Idx].Stop;
    FPoolTimers.PoolTimers[Idx].Terminate;
  end;
end;

procedure TTimerPool.RemoveAll;
var
  I: Integer;
begin
  for I := 0 to FPoolTimers.Count-1 do
  begin
    FPoolTimers.PoolTimers[I].Stop;
    FPoolTimers.PoolTimers[I].Terminate;
  end;
end;

procedure TTimerPool.Restart(AHandle: Integer);
var
  Idx: Integer;
begin
  FPoolTimerKey.Handle := AHandle;
  if FPoolTimers.Find(FPoolTimerKey, Idx) then
    FPoolTimers.PoolTimers[Idx].Restart;
end;

procedure TTimerPool.TimerTimer(Sender: TObject);
var
  I: Integer;
  PoolTimer: TPoolTimer;
begin
  I := FPoolTimers.Count-1;
  while I >= 0 do
  begin
    PoolTimer := FPoolTimers.PoolTimers[I];
  
    if PoolTimer.Terminated then
      FPoolTimers.Delete(I)
    else
      PoolTimer.FireOnExpire;
  
    Dec(I);
    if I > FPoolTimers.Count-1 then I := FPoolTimers.Count-1;
  end;
end;

{
***************************** TTimerPoolComponent *****************************
}
constructor TTimerPoolComponent.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FHandle := -1;
  FInterval := 1000;
  FEnabled := False;
end;

destructor TTimerPoolComponent.Destroy;
begin
  if FHandle <> -1 then TimerPool.Remove(FHandle);
  inherited Destroy;
end;

function TTimerPoolComponent.GetTimerPool: TTimerPool;
begin
  Result := Timers.TimerPool;
end;

procedure TTimerPoolComponent.SetEnabled(Value: Boolean);
begin
  FEnabled := Value;
  UpdateTimer;
end;

procedure TTimerPoolComponent.SetInterval(Value: LongInt);
begin
  FInterval := Value;
  UpdateTimer;
end;

procedure TTimerPoolComponent.Timer;
begin
  if Assigned(FOnTimer) then FOnTimer(Self);
end;

procedure TTimerPoolComponent.TimerPoolTrigger(Sender: TObject; Handle: 
    Integer);
begin
  if FHandle = Handle then
    Timer
  else
    raise Exception.Create('Timer pool handles do not match');
end;

{
******************************** TIntervalTimer ********************************
}
procedure TIntervalTimer.UpdateTimer;
begin
  if Handle <> -1 then TimerPool.Remove(Handle);
  Handle := -1;
  if Enabled then
    Handle := TimerPool.Add(TimerPoolTrigger, Interval);
end;

{
******************************** TOneShotTimer ********************************
}
procedure TOneShotTimer.UpdateTimer;
begin
  if Handle <> -1 then TimerPool.Remove(Handle);
  Handle := -1;
  if Enabled then
    Handle := TimerPool.AddOneTime(TimerPoolTrigger, Interval);
end;

{
********************************* TDelayTimer *********************************
}
procedure TDelayTimer.UpdateTimer;
begin
  if Handle <> -1 then TimerPool.Remove(Handle);
  Handle := -1;
  if Enabled then
    Handle := TimerPool.AddDelay(TimerPoolTrigger, Interval);
end;


function UTrunc(aDouble: Double): LongInt;
begin
  if aDouble > MaxLongInt then
    Result := Trunc( aDouble - MaxLongInt) - MaxLongInt - 2
  else
    Result := Trunc( aDouble);
end;

function ULong(L: LongInt): Double;
begin
  if L < 0 then Result := 4294967296.0 + L else Result := L;
end;
  
function AccurateTimeMS: LongInt;
begin
  Result := GetTickCount;
end;

function ElapsedTimeMS(Start, Stop : LongInt): Double;
var
  RStart, RStop: Double;
begin
  RStart := ULong(Start);
  RStop :=  ULong(Stop);
  if RStop < RStart then
    Result := 4294967296.0 + RStop - RStart
  else
    Result := RStop - RStart;
end;

function EncodeTimeMS(Hour, Min, Sec, MSec: Word): LongInt;
begin
  Result := Hour;
  Result := Result * 60 + Min;
  Result := Result * 60 + Sec;
  Result := Result * 1000 + MSec;
end;

procedure DecodeTimeMS(TimeMS: LongInt; var Hour, Min, Sec, MSec: Word);
begin
  MSec := TimeMS mod 1000;
  TimeMS := TimeMS div 1000;
  Sec := TimeMS mod 60;
  TimeMS := TimeMS div 60;
  Min := TimeMS mod 60;
  TimeMS := TimeMS div 60;
  Hour := TimeMS;
end;

procedure ExitProcedure; far;
begin
  TimerPool.Free;
end;

initialization
  RegisterStreamable(TPoolTimerCollection);
  { Create a TimerPool. }
  TimerPool := TTimerPool.Create;
  AddExitProc( ExitProcedure);
end.
